
/********************************************************************

		miscMath.c

   Miscellaneous math functions not supplied by the unix system.


*******************************************************************/

#include <stdio.h>
#include "miscMath.h"

extern char *malloc();


/*
		solveLineCartesian
   
   Determine the slope and intercept for a line that passes through
   the given points.  Put the resulting line solution in
   the structure that is pointed to by the last argument.
*/

void solveLineCartesian( point *point1, point *point2, line *theLine )

{
   /* save the points for future use */

   if( theLine->p1 == NULL )
      theLine->p1 = allocPoint();

   *theLine->p1 = *point1;

   if( theLine->p2 == NULL )
      theLine->p2 = allocPoint();

   *theLine->p2 = *point2;

   if( point1->x != point2->x ){	/* be sure line is not vertical! */

      /* slope intercept form */

      theLine->slope = ( point2->y - point1->y ) / ( point2->x - point1->x );
      theLine->intercept = point1->y - ( point1->x * theLine->slope );

      /* standard form */

      if( theLine->slope != 0 ){

	 theLine->A = 1;
	 theLine->B = -( 1 / theLine->slope );
	 theLine->C = ( theLine->p1->y / theLine->slope ) - theLine->p1->x;
      }

      else{

	 theLine->A = 0;
	 theLine->B = 1;
	 theLine->C = -point1->y;
      }
   }

   else{

      /* slope intercept form */

      theLine->slope = HUGE_VAL;		/* infinity */
      theLine->intercept = point1->x;

      /* standard form */

      theLine->A = 1;
      theLine->B = 0;
      theLine->C = -point1->x;
   }
}


/*
		solveLinePointSlope
   
   Fine the equation of a line that passes through the given
   point with the given slope.  Put the resulting line solution in
   the structure that is pointed to by the last argument.
*/

void solveLinePointSlope( point *point1, double slope, line *theLine )

{
   /* save the point for future use */

   if( theLine->p1 == NULL )
      theLine->p1 = allocPoint();

   *theLine->p1 = *point1;

   /* clear the second point since it is now undefined */

   if( theLine->p2 != NULL ){
      freePoint( theLine->p2 );
      theLine->p2 = NULL;
    }

   theLine->slope = slope;

   if( !isinf( theLine->slope ) ){	/* be sure line is not vertical! */

      /* slope intercept form */

      theLine->slope = slope;
      theLine->intercept = point1->y - ( point1->x * theLine->slope );

      /* standard form */

      if( theLine->slope != 0 ){

	 theLine->A = 1;
	 theLine->B = -( 1 / theLine->slope );
	 theLine->C = ( theLine->p1->y / theLine->slope ) - theLine->p1->x;
      }

      else{

	 theLine->A = 0;
	 theLine->B = 1;
	 theLine->C = -point1->y;
      }
   }

   else{

      /* slope intercept form */

      theLine->intercept = point1->x;

      /* standard form */

      theLine->A = 1;
      theLine->B = 0;
      theLine->C = -point1->x;
   }
}


/*
		intersectionPoint
   
   Determine the location of the intersection of the lines
   given in the arguments, and return the x coordinate of
   that point.
*/

double intersectionPoint( line *line1, line *line2 )


{
   return( ( line2->intercept - line1->intercept ) /
	   ( line1->slope - line2->slope ) );
}


/*
		evalLine

   Using the supplied value and the the flag to determine
   the axis to which it applies, compute the the corresponding
   point on the other axis.  If no corresponding point exists,
   set both values in the return point to NaN.
*/

void evalLine( line *theLine, double value, int axisFlag, point *thePoint )

{

   long dummy = 0;	/* NaN generation requires an *unused* arg */
   
   /* Be sure the line is not a point! */

   if( ( theLine->A != 0 ) || ( theLine->B != 0 ) )

      if( axisFlag == XFLAG )
	 if( theLine->B != 0 ){

	    thePoint->x = value;

	    if( theLine->A != 0 )
	       thePoint->y = theLine->slope * value  + theLine->intercept;

	    else /* line is horizontal */
	       thePoint->y = -(theLine->C/theLine->B);
	 }

	 else{ /* screw-up: x supplied for vertical line */
	    thePoint->x = signaling_nan(dummy);
	    thePoint->y = quiet_nan(dummy);  /* no point in sending two! */
	 }

      else{ /* axisFlag == YFLAG */
	 if( theLine->A != 0 ){

	    thePoint->y = value;

	    if( theLine->B != 0 )
	       thePoint->x = ( 1 / theLine->slope ) * value  -
	          	       theLine->C / theLine->A;

	    else /* line is vertical */
	       thePoint->x = -(theLine->C/theLine->A);
	 }
	 else{ /* screw-up: y supplied for horizontal line */
	    thePoint->x = signaling_nan(dummy);
	    thePoint->y = quiet_nan(dummy);
	 }
      }

   else{  /* major screw-up:  caller sent an undefined line */
      thePoint->x = signaling_nan(dummy);
      thePoint->y = quiet_nan(dummy);
   }
}

/*
		evalLineFx
   Compute and return the y value of the line specified by the
   arguments at the point specified by the arguments.
*/

double evalLineFx( double x, line *theLine )


{
   return( ( theLine->slope * x ) + theLine->intercept );
}

/*
		linesIntersect

   Determine if the two line segments intersect.
   (From Sedgewick, Elementary Geometric Methods.)
*/

int linesIntersect( line *line1, line *line2 )


{
   return( ( sameSide( line1, line2->p1, line2->p2 ) <= 0 ) &&
           ( sameSide( line2, line1->p1, line1->p2 ) <= 0 ) );
}


/*
		sameSide

   Determine if two points are on the same side of a line.
   Returns a negative value if the points are on opposite
   sides, a positive value if on the same side.
   (From Sedgewick, Elementary Geometric Methods.)

*/

double sameSide( line *theLine, point *point1, point *point2 )

{
   double dx;
   double dy;
   double dx1;
   double dy1;
   double dx2;
   double dy2;

   dx = theLine->p2->x - theLine->p1->x;
   dy = theLine->p2->y - theLine->p1->y;

   dx1 = point1->x - theLine->p1->x;
   dy1 = point1->y - theLine->p1->y;

   dx2 = point2->x - theLine->p2->x;
   dy2 = point2->y - theLine->p2->y;

   return( ( dx * dy1 - dy * dx1 ) * ( dx * dy2 - dy * dx2 ) );
}


/*		
		distance
   
   Calculate the cartesian distance between the two pairs of points
   given in the arguments.
*/

double distance( point *p1, point *p2 )

{
   return( sqrt( ( p1->x - p2->x ) * ( p1->x - p2->x ) +
  		 ( p1->y - p2->y ) * ( p1->y - p2->y ) ) );
}


/*
		pointLineSeparation

   Find the distance between the specified line and
   point.
*/

double pointLineSeparation( point *pt, line *line )


{
   return( abs( (line->A * pt->x) + (line->B * pt->y) + line->C  )  / 
	   sqrt( (line->A * line->A) + (line->B * line->B) ) );

}

/*
		polar

   Convert the coordinate given by x,y to polar coordinates and
   place the results into the variables pointed to by the
   last two arguments
*/


void cartesianToPolar( double x, double y, double *radius, double *angle )

{
   *radius = sqrt( ( x * x ) + ( y * y ) );
   *angle  = atan2( y, x );
}



/*
		allocLine

   Allocate an "aLine" structure and return a pointer to it.

*/

line *allocLine()

{
   struct aLine *tmpLine;

   tmpLine = (struct aLine*)malloc( (unsigned)sizeof(struct aLine) );

   tmpLine->p1 = NULL;
   tmpLine->p2 = NULL;

   return( tmpLine );
}



/*
	freeLine

   Free an "aLine" structure.
   Be sure the associated points get freed, too!
*/

void freeLine( line *theLine )

{
   if( theLine->p1 != NULL )
      freePoint( theLine->p1 );

   if( theLine->p2 != NULL )
      freePoint( theLine->p2 );

   (void)free( (char*)theLine );
}



/*
		allocPoint

   Allocate an "aPoint" structure and return a pointer to it.

*/

point *allocPoint()

{
   return( (struct aPoint*)malloc( (unsigned)sizeof( struct aPoint ) ) );
}



/*
	freePoint

   Free an "aPoint" structure.
*/

void freePoint( point *thePoint )

{
   (void)free( (char*)thePoint );
}

